package mhtree.benchmarking;

import cz.muni.fi.disa.similarityoperators.cover.AbstractRepresentation;
import messif.algorithms.AlgorithmMethodException;
import messif.buckets.BucketStorageException;
import messif.objects.LocalAbstractObject;
import messif.objects.impl.ObjectFloatVectorNeuralNetworkL2;
import messif.objects.util.AbstractObjectList;
import messif.objects.util.RankedAbstractObject;
import messif.objects.util.StreamGenericAbstractObjectIterator;
import messif.operations.Approximate;
import messif.operations.data.BulkInsertOperation;
import messif.operations.query.ApproxKNNQueryOperation;
import messif.operations.query.KNNQueryOperation;
import messif.statistics.Statistics;
import mhtree.InsertType;
import mhtree.MHTree;
import mhtree.MergingMethod;
import mhtree.ObjectToNodeDistance;
import mtree.MTree;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import messif.algorithms.Algorithm;

import static mhtree.ObjectToNodeDistance.AVERAGE;
import static mhtree.ObjectToNodeDistance.FURTHEST;
import static mhtree.ObjectToNodeDistance.NEAREST;

public class RunBenchmark {
    public static void main(String[] args) throws IOException, AlgorithmMethodException, InstantiationException, NoSuchMethodException, BucketStorageException {
        if (args.length != 7) {
            throw new IllegalArgumentException("Unexpected number of params");
        }

        boolean isMHtree = switch (args[0]) {
            case "MHULL-TREE" -> true;
            default -> false;
        };
        
        //Statistics.enableGlobally();
        Statistics.disableGlobally();
        AbstractRepresentation.PrecomputedDistances.COMPUTATION_THREADS = 16;

        List<LocalAbstractObject> objects = loadDataset(args[1]);
        int leafCapacity = Integer.parseInt(args[2]);
        int nodeDegree = Integer.parseInt(args[3]);
        List<LocalAbstractObject> queries = loadDataset(args[4]);
        
        InsertType insertType = args[5].equals("INCREMENTAL") ? InsertType.INCREMENTAL : InsertType.GREEDY;

        ObjectToNodeDistance objectToNodeDistance = switch (args[6]) {
            case "FURTHEST" -> FURTHEST;
            case "AVERAGE" -> AVERAGE;
            default -> NEAREST;
        };
        
        final MHTreeConfig cfg = new MHTreeConfig(
                leafCapacity,
                nodeDegree,
                insertType,
                objectToNodeDistance
        );
        final int[] ks = new int[]{1, 3, 5, 10, 20, 50, 100};
        if (isMHtree) {
            percentageToRecallMHTree(cfg,
                    objects,
                    Arrays.asList(NEAREST),
                    //Arrays.asList(NEAREST, FURTHEST, AVERAGE),
                    queries, ks);
        } else {
            percentageToRecallMTree(cfg,
                    objects,
                    queries, ks);
        }
    }

    private static void percentageToRecallMHTree(MHTreeConfig config, List<LocalAbstractObject> objects, 
            final List<ObjectToNodeDistance> distFuncs,
            List<LocalAbstractObject> queries,
            int[] ks) throws BucketStorageException, RuntimeException {

        long buildingStartTimeStamp = System.currentTimeMillis();

        MHTree mhTree = new MHTree.MHTreeBuilder(objects, config.leafCapacity, config.nodeDegree)
                .objectToNodeDistance(config.objectToNodeDistance)
                .mergingMethod(MergingMethod.HULL_BASED_MERGE)
                .build();

        long buildinTime = System.currentTimeMillis() - buildingStartTimeStamp;
        //OperationStatistics.getLocalThreadStatistics().printStatistics();
        mhTree.printStatistics();
        //System.out.println("Fat factor: " + mhTree.getFatFactor());
        System.out.println("Building time: " + buildinTime + " msec");

        System.out.println("kNN queries will be executed for k=" + Arrays.toString(ks));
        
        for (ObjectToNodeDistance dist : distFuncs) {
            System.gc();

            mhTree.getNodes().forEach(node -> node.objectToNodeDistance = dist);

            System.out.println("leafCapacity,nodeDegree,objectToNodeDistance,k,percentage,recall (min),recall (avg),recall (med),recall (max), time (msec)");

            Map<String, List<RankedAbstractObject>> kNNResults = prepareGroundTruth(ks, queries, mhTree);

            for (int k : ks) {
                List<ApproxKNNQueryOperation> approxOperations = queries
                        .parallelStream()
                        .map(object -> new ApproxKNNQueryOperation(object, k, 0, Approximate.LocalSearchType.PERCENTAGE, LocalAbstractObject.UNKNOWN_DISTANCE))
                        .collect(Collectors.toList());
                approxOperations
                        .parallelStream()
                        .forEach(op -> op.suppData = new SearchState(mhTree, op));

                double minimalRecall = 0;
                for (int percentage = 0; percentage <= 100; percentage += 5) {
                    final int approxLimit = Math.round((float) objects.size() * (float) percentage / 100f);
                    approxOperations
                            .parallelStream()
                            .filter(op -> ((SearchState) op.suppData).recall != 1d)
                            .forEach(op -> {
                                SearchState searchState = (SearchState) op.suppData;
                                try {
                                    //mhTree.approxKNN(op);
                                    mhTree.executeOperation(op);
                                } catch (AlgorithmMethodException | NoSuchMethodException ex) {
                                }
                                searchState.time = op.getParameter("OperationTime", Long.class);
                                searchState.recall = PerformanceMeasures.measureRecall(op, kNNResults);
                                searchState.approximateState.limit = approxLimit;
                                op.resetAnswer();
                            });

                    Stats recallStats = new Stats(
                            approxOperations
                                    .stream()
                                    .map(op -> ((SearchState) op.suppData).recall)
                                    .collect(Collectors.toList())
                    );
                    Stats timeStats = new Stats(
                            approxOperations
                                    .stream()
                                    .map(op -> (double)((SearchState) op.suppData).time)
                                    .collect(Collectors.toList())
                    );

                    System.out.println(String.join(",",
                            String.valueOf(config.leafCapacity),
                            String.valueOf(config.nodeDegree),
                            String.valueOf(dist),
                            String.valueOf(k),
                            String.valueOf(percentage),
                            String.format("%.2f,%.2f,%.2f,%.2f",
                                    recallStats.getMin(),
                                    recallStats.getAverage(),
                                    recallStats.getMedian(),
                                    recallStats.getMax()),
                            String.format("%.2f", timeStats.getAverage())));

                    minimalRecall = recallStats.getMin();
                    if (minimalRecall == 1.0)
                        break;
                }
            }
        }
    }

    private static void percentageToRecallMTree(MHTreeConfig config, List<LocalAbstractObject> objects, 
            List<LocalAbstractObject> queries,
            int[] ks) throws NoSuchMethodException, AlgorithmMethodException, RuntimeException, InstantiationException {

        long buildingStartTimeStamp = System.currentTimeMillis();

        MTree mTree = new MTree(config.nodeDegree, config.leafCapacity);

        Collections.shuffle(objects);

        BulkInsertOperation opIns = new BulkInsertOperation(objects);

        mTree.insert(opIns);
        long buildingTime = System.currentTimeMillis() - buildingStartTimeStamp;

        mTree.printStatistics();
        //System.out.println("Fat factor: " + mTree.getFatFactor());
        System.out.println("Building time: " + buildingTime + " msec");

        System.out.println("kNN queries will be executed for k=" + Arrays.toString(ks));
        
        System.out.println("leafCapacity,nodeDegree,objectToNodeDistance,k,percentage,recall (min),recall (avg),recall (med),recall (max), time (msec)");

        Map<String, List<RankedAbstractObject>> kNNResults = prepareGroundTruth(ks, queries, mTree);
        

//        int numberOfQueries = queries.size();
        for (int k : ks) {
            double minimalRecall = 0;
            for (int percentage = 0; percentage <= 100; percentage += 5) {
                final int approxLimit = percentage;
                List<ApproxKNNQueryOperation> approxOperations = queries
                        .parallelStream()
                        .map(object -> new ApproxKNNQueryOperation(object, k, approxLimit, Approximate.LocalSearchType.PERCENTAGE, LocalAbstractObject.UNKNOWN_DISTANCE))
                        .collect(Collectors.toList());
                approxOperations
                        .parallelStream()
                        .forEach(op -> op.suppData = new SearchInfoState());
                
                approxOperations
                        .parallelStream()
                        .forEach(op -> {
                            SearchInfoState searchState = (SearchInfoState) op.suppData;
                            try {
                                //mhTree.approxKNN(op);
                                mTree.executeOperation(op);
                            } catch (AlgorithmMethodException | NoSuchMethodException ex) {
                            }
                            searchState.time = op.getParameter("OperationTime", Long.class);
                            searchState.recall = PerformanceMeasures.measureRecall(op, kNNResults);
                            //searchState.approximateState.limit += Math.round((float) objects.size() * (float) percentageStep / 100f);
                        });

                Stats recallStats = new Stats(
                        approxOperations
                                .stream()
                                .map(op -> ((SearchInfoState) op.suppData).recall)
                                .collect(Collectors.toList())
                );
                Stats timeStats = new Stats(
                        approxOperations
                                .stream()
                                .map(op -> (double)((SearchInfoState) op.suppData).time)
                                .collect(Collectors.toList())
                );
                
//                for (int i = 0; i < numberOfQueries; i++) {
//                    if (recalls.get(i) != 1.0) {
//                        ApproxKNNQueryOperation operation = new ApproxKNNQueryOperation(
//                                queries.get(i),
//                                k,
//                                percentage,
//                                Approximate.LocalSearchType.PERCENTAGE,
//                                LocalAbstractObject.UNKNOWN_DISTANCE
//                        );
//
//                        mTree.executeOperation(operation);
//
//                        times.set(i, (double)operation.getParameter("OperationTime", Long.class));
//                        recalls.set(i, PerformanceMeasures.measureRecall(operation, mTree));
//                    }
//                }

                System.out.println(String.join(",",
                        String.valueOf(config.leafCapacity),
                        String.valueOf(config.nodeDegree),
                        "",
                        String.valueOf(k),
                        String.valueOf(percentage),
                        String.format("%.2f,%.2f,%.2f,%.2f",
                                recallStats.getMin(),
                                recallStats.getAverage(),
                                recallStats.getMedian(),
                                recallStats.getMax()),
                        String.format("%.2f", timeStats.getAverage())));

                minimalRecall = recallStats.getMin();
                if (minimalRecall == 1.0)
                    break;
            }
        }
    }

    private static Map<String, List<RankedAbstractObject>> prepareGroundTruth(int[] ks, List<LocalAbstractObject> queries, Algorithm alg) {
        int maxK = Arrays.stream(ks).max().getAsInt();
        
        List<KNNQueryOperation> kNNOperations = queries
                .parallelStream()
                .map(object -> new KNNQueryOperation(object, maxK))
                .collect(Collectors.toList());
        Map<String, List<RankedAbstractObject>> kNNResults = kNNOperations
                .parallelStream()
                .map((KNNQueryOperation op) -> {
                    try {
                        alg.executeOperation(op);
                    } catch (AlgorithmMethodException | NoSuchMethodException ex) { }
                    KnnResultPair pair = new KnnResultPair();
                    pair.id = op.getQueryObject().getLocatorURI();
                    pair.kNNObjects = new ArrayList<>(maxK);
                    for (RankedAbstractObject o : op)
                        pair.kNNObjects.add(o);
                    return pair;
                })
                .collect(Collectors.toMap(entry -> entry.id, entry -> entry.kNNObjects));
        return kNNResults;
    }

    private static List<LocalAbstractObject> loadDataset(String path) throws IOException {
        return new AbstractObjectList<>(new StreamGenericAbstractObjectIterator<>(ObjectFloatVectorNeuralNetworkL2.class, path));
    }
}
